import re
import time
import os
import json
from pathlib import Path

from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import HttpResponse, JsonResponse
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from django.utils import timezone

from course_flow import models

owned_throughmodels = [
    "node",
    "nodeweek",
    "week",
    "weekworkflow",
    "workflow",
    "workflowproject",
    "project",
    "columnworkflow",
    "workflow",
    "workflowproject",
    "project",
    "outcome",
    "outcomeoutcome",
    "outcome",
]


default_column_settings = {
    "0":{"colour":"#6738ff","icon":"other"},
    "1":{"colour":"#0b118a","icon":"ooci"},
    "2":{"colour":"#114cd4","icon":"home"},
    "3":{"colour":"#11b3d4","icon":"instruct"},
    "4":{"colour":"#04d07d","icon":"students"},
    "10":{"colour":"#6738ff","icon":"other"},
    "11":{"colour":"#ad351d","icon":"homework"},
    "12":{"colour":"#ed4a28","icon":"lesson"},
    "13":{"colour":"#ed8934","icon":"artifact"},
    "14":{"colour":"#f7ba2a","icon":"assessment"},
    "20":{"colour":"#369934","icon":"other"}
}


def get_alphanum(string):
    if(string is None):return ""
    return re.sub(r"\W+", "", string)


# Create a regex from dict keys
def multiple_replace(dict, text):
    regex = re.compile("(%s)" % "|".join(map(re.escape, dict.keys())))
    return regex.sub(lambda mo: dict[mo.string[mo.start() : mo.end()]], text)


def dateTimeFormat():
    return "%Y/%m/%d"


def dateTimeFormatNoSpace():
    return "%Y_%m_%d_%H_%m_%s"


def get_model_from_str(model_str: str):
    return ContentType.objects.get(model=model_str).model_class()


def get_parent_model_str(model_str: str) -> str:

    return owned_throughmodels[owned_throughmodels.index(model_str) + 1]


def get_parent_model(model_str: str):

    return ContentType.objects.get(
        model=get_parent_model_str(model_str)
    ).model_class()


def linkIDMap(link):
    return link.id


def get_descendant_outcomes(outcome):
    return models.Outcome.objects.filter(
        Q(parent_outcomes=outcome)
        | Q(parent_outcomes__parent_outcomes=outcome)
    )

def get_all_outcomes_for_outcome(outcome):
    outcomes = models.Outcome.objects.filter(
        Q(parent_outcomes=outcome)
        | Q(parent_outcomes__parent_outcomes=outcome)
    ).prefetch_related(
        "outcome_horizontal_links", "child_outcome_links", "sets"
    )
    outcomeoutcomes = models.OutcomeOutcome.objects.filter(
        Q(parent=outcome) | Q(parent__parent_outcomes=outcome)
    )
    return outcomes, outcomeoutcomes


def get_all_outcomes_for_workflow(workflow):
    outcomes = models.Outcome.objects.filter(
        Q(workflow=workflow)
        | Q(parent_outcomes__workflow=workflow)
        | Q(parent_outcomes__parent_outcomes__workflow=workflow)
    ).prefetch_related(
        "outcome_horizontal_links", "child_outcome_links", "sets"
    )
    outcomeoutcomes = models.OutcomeOutcome.objects.filter(
        Q(parent__workflow=workflow)
        | Q(parent__parent_outcomes__workflow=workflow)
    )
    return outcomes, outcomeoutcomes


def get_all_outcomes_ordered_for_outcome(outcome):
    outcomes = [outcome]
    for outcomeoutcome in outcome.child_outcome_links.filter(
        child__deleted=False
    ).order_by("rank"):
        outcomes += get_all_outcomes_ordered_for_outcome(outcomeoutcome.child)
    return outcomes


def get_all_outcomes_ordered(workflow):
    outcomes = []
    for outcomeworkflow in workflow.outcomeworkflow_set.filter(
        outcome__deleted=False
    ).order_by("rank"):
        outcomes += get_all_outcomes_ordered_for_outcome(
            outcomeworkflow.outcome
        )
    return outcomes

def get_direct_children_of_outcome_ordered(outcome):
    return(models.Outcome.objects.filter(
        parent_outcomes=outcome,deleted=False
    ).order_by("parent_outcome_links__rank"))



def get_base_outcomes_ordered_filtered(workflow, extra_filter=Q()):
    return (
        models.Outcome.objects.filter(workflow=workflow, deleted=False)
        .filter(extra_filter)
        .order_by("outcomeworkflow__rank")
        .distinct()
    )


def get_all_outcomes_ordered_filtered(workflow, extra_filter):
    outcomes = []
    for outcome in (
        models.Outcome.objects.filter(workflow=workflow, deleted=False)
        .filter(extra_filter)
        .order_by("outcomeworkflow__rank")
        .distinct()
    ):
        outcomes += get_all_outcomes_ordered_for_outcome(outcome)
    return outcomes

def get_unique_outcomenodes(node):
    return (
        node.outcomenode_set.exclude(
            Q(outcome__deleted=True)
            | Q(outcome__parent_outcomes__deleted=True)
            | Q(outcome__parent_outcomes__parent_outcomes__deleted=True)
        )
        .exclude(outcome__parent_outcomes__node=node)
        .exclude(outcome__parent_outcomes__parent_outcomes__node=node)
        .order_by(
            "outcome__parent_outcome_links__parent__parent_outcome_links__parent__outcomeworkflow__rank",
            "outcome__parent_outcome_links__parent__outcomeworkflow__rank",
            "outcome__outcomeworkflow__rank",
            "outcome__parent_outcome_links__parent__parent_outcome_links__rank",
            "outcome__parent_outcome_links__rank",
        )
    )

def get_outcomenodes(node):
    return node.outcomenode_set.exclude(
        Q(outcome__deleted=True)
        | Q(outcome__parent_outcomes__deleted=True)
        | Q(outcome__parent_outcomes__parent_outcomes__deleted=True)
    ).order_by(
        "outcome__parent_outcome_links__parent__parent_outcome_links__parent__outcomeworkflow__rank",
        "outcome__parent_outcome_links__parent__outcomeworkflow__rank",
        "outcome__outcomeworkflow__rank",
        "outcome__parent_outcome_links__parent__parent_outcome_links__rank",
        "outcome__parent_outcome_links__rank",
    )

#From an outcomenode, create a Q to get the outcome and all its parents, heirarchically organized
def get_outcomenode_trace(ocn):
    oc = ocn.outcome
    return models.Outcome.objects.filter(
        Q(pk=oc.pk)
        | Q(children=oc)
        | Q(children__children=oc)
        | Q(children__children__children=oc)
    ).distinct().order_by("depth")

def get_unique_outcomehorizontallinks(outcome):
    return (
        outcome.outcome_horizontal_links.exclude(
            Q(parent_outcome__deleted=True)
            | Q(parent_outcome__parent_outcomes__deleted=True)
            | Q(parent_outcome__parent_outcomes__parent_outcomes__deleted=True)
        )
        .exclude(
            parent_outcome__parent_outcomes__reverse_horizontal_outcomes=outcome
        )
        .exclude(
            parent_outcome__parent_outcomes__parent_outcomes__reverse_horizontal_outcomes=outcome
        )
        .order_by(
            "parent_outcome__parent_outcome_links__parent__parent_outcome_links__parent__outcomeworkflow__rank",
            "parent_outcome__parent_outcome_links__parent__outcomeworkflow__rank",
            "parent_outcome__outcomeworkflow__rank",
            "parent_outcome__parent_outcome_links__parent__parent_outcome_links__rank",
            "parent_outcome__parent_outcome_links__rank",
        )
    )


def get_parent_nodes_for_workflow(workflow):
    nodes = (
        models.Node.objects.filter(linked_workflow=workflow)
        .exclude(
            Q(deleted=True)
            | Q(week__deleted=True)
            | Q(week__workflow__deleted=True)
        )
        .prefetch_related("outcomenode_set")
    )
    return nodes


def get_nondeleted_favourites(user):
    return list(
        models.Project.objects.filter(favourited_by__user=user)
    ) + list(models.Workflow.objects.filter(favourited_by__user=user))

    # return models.Favourite.objects.filter(user=user).exclude(
    #     Q(
    #         object_id__in=models.Workflow.objects.filter(
    #             Q(deleted=True) | Q(project__deleted=True)
    #         ),
    #         content_type=ContentType.objects.get_for_model(models.Workflow)
    #     )
    #     | Q(
    #         object_id__in=models.Project.objects.filter(deleted=True),
    #         content_type=ContentType.objects.get_for_model(models.Project)
    #     )
    # )


def check_possible_parent(workflow, parent_workflow, same_project):
    order = ["activity", "course", "program"]
    try:
        if order.index(workflow.type) == order.index(parent_workflow.type) - 1:
            if same_project:
                if workflow.get_project() == parent_workflow.get_project():
                    return True
            else:
                return True
    except IndexError:
        pass
    return False


def get_classrooms_for_student(user):
    return models.Project.objects.filter(
        liveproject__liveprojectuser__user=user,
        deleted=False,
    )


def get_user_permission(obj, user):
    if obj.type in ["workflow", "course", "activity", "program"]:
        obj = models.Workflow.objects.get(pk=obj.pk)

    if user is None or not user.is_authenticated:
        return models.ObjectPermission.PERMISSION_NONE
    if obj.author == user:
        return models.ObjectPermission.PERMISSION_EDIT
    permissions = models.ObjectPermission.objects.filter(
        user=user,
        content_type=ContentType.objects.get_for_model(obj),
        object_id=obj.id,
    )
    if permissions.count() == 0:
        return models.ObjectPermission.PERMISSION_NONE
    return permissions.first().permission_type


def get_user_role(obj, user):
    if user is None or not user.is_authenticated:
        return models.LiveProjectUser.ROLE_NONE
    if obj.type == "liveproject":
        liveproject = obj
        project = obj.project
    elif obj.type == "project":
        try:
            liveproject = obj.liveproject
            project = obj
        except AttributeError:
            return models.LiveProjectUser.ROLE_NONE
    elif obj.is_strategy:
        project = None
        liveproject = None
    else:
        try:
            project = obj.get_project()
            liveproject = project.liveproject
        except AttributeError:
            return models.LiveProjectUser.ROLE_NONE
    if liveproject is None:
        return models.LiveProjectUser.ROLE_NONE
    if hasattr(obj, "author") and obj.author == user:
        return models.LiveProjectUser.ROLE_TEACHER
    permissions = models.LiveProjectUser.objects.filter(
        user=user, liveproject=liveproject
    )
    if permissions.count() == 0:
        return models.LiveProjectUser.ROLE_NONE
    return permissions.first().role_type


def user_workflow_url(workflow, user):
    user_permission = get_user_permission(workflow, user)
    user_role = get_user_role(workflow, user)
    can_view = False
    is_public = workflow.public_view
    if user is not None and user.is_authenticated and workflow.published:
        if Group.objects.get(name=settings.TEACHER_GROUP) in user.groups.all():
            can_view = True
    if user_permission != models.ObjectPermission.PERMISSION_NONE:
        can_view = True
    if user_role != models.LiveProjectUser.ROLE_NONE:
        can_view = True
    if can_view:
        return reverse(
            "course_flow:workflow-update", kwargs={"pk": workflow.pk}
        )
    if is_public:
        return reverse(
            "course_flow:workflow-public", kwargs={"pk": workflow.pk}
        )
    if user is None or not user.is_authenticated:
        return "nouser"
    return "noaccess"


def save_serializer(serializer) -> HttpResponse:
    if serializer:
        if serializer.is_valid():
            serializer.save()
            return JsonResponse({"action": "posted"})
        else:
            return JsonResponse({"action": "error"})
    else:
        return JsonResponse({"action": "error"})


def get_relevance(obj, name_filter, keywords):
    if obj.title is None:
        title = ""
    else:
        title = obj.title.lower()
    if obj.description is None:
        description = ""
    else:
        description = obj.description.lower()
    if obj.author is None:
        first = ""
        last = ""
        username = ""
    else:
        if obj.author.first_name is None:
            first = ""
        else:
            first = obj.author.first_name
        if obj.author.last_name is None:
            last = ""
        else:
            last = obj.author.last_name
        if obj.author.username is None:
            username = ""
        else:
            username = obj.author.username
    relevance = ""
    to_check = [name_filter] + keywords
    keys = [title, last, first, username, description]
    for key in keys:
        for keyword in to_check:
            if keyword == "":
                continue
            if key.startswith(keyword):
                relevance += "0"
            elif key.find(" " + keyword) >= 0:
                relevance += "1"
            else:
                relevance += "2"
    return relevance


def benchmark(identifier, last_time):
    current_time = time.time()
    print("Completed " + identifier + " in " + str(current_time - last_time))
    return current_time


def clean_old_exports(user_dir, max_jobs: int = 2):
    user_dir = Path(user_dir)
    if not user_dir.exists():
        return

    # Find all job JSON files
    job_files = list(user_dir.glob('*.json'))
    jobs = []

    for job_file in job_files:
        try:
            with open(job_file) as f:
                job_data = json.load(f)
                created = job_data.get('created')
                if created:
                    dt = parse_datetime(created)
                    jobs.append((dt, job_data['filename'], job_file))
        except Exception as e:
            # If the file is corrupt, we delete it
            try:
                print("deleting a corrupt job file")
                print(e)
                job_file.unlink()
            except Exception as delete_err:
                print(f"Error deleting corrupt file: {delete_err}")

    # Sort by creation date
    jobs.sort(key=lambda tup: tup[0])  # oldest first

    # If too many, delete oldest
    while len(jobs) >= max_jobs + 1:
        _, filename, job_file = jobs.pop(0)
        job_file.unlink()
        if filename is not None:
            data_file = user_dir / filename
            if data_file.exists():
                data_file.unlink()

    valid_filenames = {filename for _, filename, _ in jobs}

    # Delete any files in the folder that aren't .json and not in the valid set
    for f in user_dir.iterdir():
        if f.suffix in {'.csv', '.xlsx'} and f.name not in valid_filenames:
            try:
                f.unlink()
                print(f"Deleted orphaned export file: {f}")
            except Exception as e:
                print(f"Error deleting orphaned file {f}: {e}")

    #Delete any files older than 24 hours
    EXPORT_MAX_AGE_SECONDS = 86400

    now = timezone.now()
    for f in user_dir.iterdir():
        try:
            mtime = timezone.datetime.fromtimestamp(f.stat().st_mtime, tz=timezone.utc)
            if (now - mtime).total_seconds() > EXPORT_MAX_AGE_SECONDS:
                f.unlink()
        except Exception as e:
            print(f"Error deleting {f}: {e}")


from io import BytesIO
import time
import traceback
import numpy as np

import pandas as pd
from openpyxl.styles import PatternFill
from django.db.models import Q
from django.utils.translation import gettext as _
from django.utils import timezone

from course_flow import analytics, models

from course_flow.export_functions import allowed_sets_Q

from .models import (
    Course,
    Node,
    Outcome,
    OutcomeNode,
    OutcomeWorkflow,
    Program,
    Week,
    WeekWorkflow,
    Column,
)
from .serializers import (
    NodeExportSerializerForFormatted,
    WeekExportSerializer,
    WorkflowExportSerializer,
    WorkflowExportSerializerWithPonderation,
    ColumnExportSerializer,
)
from .utils import (
    get_alphanum,
    dateTimeFormat,
    benchmark,
)

def hex_to_rgb(hex_color):
    hex_color = hex_color.lstrip("#")
    return tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))


def luminance(rgb):
    r, g, b = rgb
    return 0.299 * r + 0.587 * g + 0.114 * b


def readable_text_color(bg_hex, threshold=140):
    """
    Returns '#000000' or '#FFFFFF' depending on background brightness.
    """
    lum = luminance(hex_to_rgb(bg_hex))
    return "#000000" if lum > 140 else "#FFFFFF"


def get_workflow_full_table(workflow, allowed_sets):
    entries = []
    for week in Week.objects.filter(workflow=workflow, deleted=False).order_by(
        "weekworkflow__rank"
    ):
        week_serialized = WeekExportSerializer(week).data
        week_serialized["week"]=week_serialized["title"]
        del week_serialized["title"]
        entries += [week_serialized]
        nodes = NodeExportSerializerForFormatted(
            Node.objects.filter(week=week, deleted=False)
            .filter(allowed_sets_Q(allowed_sets))
            .distinct()
            .order_by("nodeweek__rank"),
            many=True,
        ).data
        entries += nodes
    #Create the original dataframe
    df = pd.DataFrame(
        entries, columns=["type", "week","title", "description", "outcomes", "column_order", "id","time_required","time_units_display"]
    )
    #Add the time to the titles
    mask = df["time_required"].notna() & (df["time_required"] != 0)
    sep = np.where(df["time_units_display"].astype(str).str.len() > 0, " ", "")
    df["title"] = df["title"].where(
        ~mask,
        df["title"] + " (" +
        df["time_required"].astype(str) + sep +
        df["time_units_display"] + ")"
    )
    df = df.drop(columns=["time_required","time_units_display"])
    #Since we have weeks and nodes, we'll need a unique id for our objects and we want to preserve this order for later
    df["uid"] = df.index
    df["description"] = df["description"].fillna("\n")
    df["column_order"] = df["column_order"].fillna(-1)
    df["column_order"] = df["column_order"].astype(int)

    #Get the column colour info
    df_col = pd.DataFrame(
        ColumnExportSerializer(
            Column.objects.filter(workflow=workflow, deleted=False).order_by("columnworkflow__rank"),
            many=True
        ).data, 
        columns=["title", "colour"]
    )
    df_col = df_col.rename(columns={"title":"column_title"})
    df_col["column_order"]=df_col.index
    #Merge it into our dataframe
    df = df.merge(df_col,on="column_order",how="left")

    #Pivot our tables to get a sparse array that mimics the column structure of workflows
    wide_title = (
        df[df["type"]=="node"].pivot(index="uid",columns="column_order", values="title")
    )
    wide_title["rowtype"]="a_title"
    wide_description = (
        df[df["type"]=="node"].pivot(index="uid",columns="column_order", values="description")
    )
    wide_description["rowtype"]="b_description"
    #Create a table with titles and descriptions in separate rows
    interleaved = (pd.concat([wide_title,wide_description])
        .reset_index()
    )
    pivoted = df.merge(interleaved, on=["uid"],how="outer").sort_values(["uid","rowtype"])
    pivoted = pivoted.drop(columns=["title","description","column_order","id","uid","column_title"])
    pivoted.loc[pivoted["type"]=="week","rowtype"]="week"

    pivoted.loc[pivoted["rowtype"]=="a_title","outcomes"]=""


    #Add a row at the top for the column
    col_row = df_col.drop(columns=["colour","column_order"]).T
    col_row["week"]=_("Week")
    col_row["outcomes"]=_("Outcomes")
    col_row["rowtype"]=["columns"]

    #Add a few lines for the workflow info
    wf_serialized = WorkflowExportSerializerWithPonderation(workflow).data
    wf_info = [{
        "week":wf_serialized.get("title"),
        "rowtype":"wf_title",
    },{
        "week":_("Description"),
        "outcomes":wf_serialized.get("description"),
        "rowtype":"wf_description",
    },{
        "week":f'{_("Ponderation")} ({_("Theory")}/{_("Practical")}/{_("Individual")})',
        "outcomes": f'{wf_serialized.get("ponderation_theory")}/{wf_serialized.get("ponderation_practical")}/{wf_serialized.get("ponderation_individual")}',
        "rowtype":"wf_ponderation",
    }]

    df_info = pd.DataFrame(wf_info)

    with_columns = pd.concat([df_info,col_row.reset_index(),pivoted]).reindex(columns=pivoted.columns)

    with_columns = with_columns.drop(columns=["type"])

    #move colour to the end
    with_columns["colour"] = with_columns.pop("colour")

    pd.set_option("display.max_colwidth", None)
    return (with_columns.reset_index().drop(columns=["index"]),df_col)



def get_workflows_export(model_object, object_type, export_format, allowed_sets):
    if object_type == "project":
        workflows = list(model_object.workflows.filter(deleted=False))
    else:
        workflows = [model_object]
    with BytesIO() as b:
        if export_format == "excel":
            title_colour = "#04BA74"
            week_colour = "#B7DEE8"
            week_header_colour = "#92CDDC"
            outcome_header_colour = "#B1A0C7"
            with pd.ExcelWriter(b, engine="xlsxwriter") as writer:
                workbook = writer.book 
                wrap_format = workbook.add_format()
                wrap_format.set_text_wrap()
                wrap_format.set_align("top")

                title_format = workbook.add_format({
                    "bg_color": title_colour,
                    "font_color": "#FFFFFF",
                    "font_size": 18,
                    "bold": True,
                    "text_wrap": True,
                    "valign": "top",
                })
                week_format = workbook.add_format({
                    "bg_color": week_colour,
                    "font_size":14,
                    "bold": True,
                    "text_wrap": True,
                    "valign": "top",
                })
                week_header_format = workbook.add_format({
                    "bg_color": week_header_colour,
                    "font_size": 16,
                    "bold": True,
                    "text_wrap": True,
                    "valign": "top",
                })
                outcome_header_format = workbook.add_format({
                    "bg_color": outcome_header_colour,
                    "font_size": 16,
                    "bold": True,
                    "text_wrap": True,
                    "valign": "top",
                })
                ponderation_format = workbook.add_format({
                    "font_size": 16,
                    "bold": True,
                    "text_wrap": True,
                    "valign": "top",
                })
                description_text_format = workbook.add_format({
                    "valign": "top",
                    "text_wrap": True,
                })

                for workflow in workflows:
                    (df, columns) = get_workflow_full_table(workflow, allowed_sets)
                    sheet_name = (
                        get_alphanum(workflow.title) + "_" + str(workflow.pk)
                    )[:30]
                    #Remove some columns when writing to excel
                    df.drop(columns=["rowtype","colour"]).to_excel(
                        writer,
                        sheet_name=sheet_name,
                        index=False,
                        header=False,
                    )
                    ws = writer.sheets[sheet_name]
                    #Set column widths
                    ws.set_column(0, 20, 40, wrap_format)
                    #Sticky header
                    ws.freeze_panes(4,0)
                    header = list(df.columns)
                    #Get our data columns that have to be coloured in
                    #The key is the original columns_order (the dataframe column), value is the column in excel
                    data_cols = {}
                    for i, name in enumerate(header):
                        if str(name).isdigit():
                            data_cols[name] = i

                    format_cache = {}
                    for row_idx, row in columns.iterrows():
                        colour = row["colour"]
                        column_order = row["column_order"]
                        if column_order in data_cols.keys():
                            format_cache[data_cols[column_order]] = workbook.add_format({
                                "bg_color": colour,
                                "border": 1,
                                "font_size": 16,
                                "bold": True,
                                "font_color": readable_text_color(colour),
                                "text_wrap": True,
                                "valign": "top",
                            })

                    for row_idx, row in df.iterrows():
                        rowtype = row["rowtype"]
                        if pd.isna(rowtype) or rowtype == "":
                            continue
                        if rowtype == "wf_title":
                            ws.set_row(row_idx,None,title_format)
                        elif rowtype == "week":
                            ws.set_row(row_idx,None,week_format)
                        elif rowtype == "columns":
                            ws.set_row(row_idx,None,week_header_format)
                            outcome_col = df.columns.get_loc("outcomes")
                            ws.write(row_idx,outcome_col,row["outcomes"],outcome_header_format)
                            for col_idx in data_cols.values():
                                value = row.iloc[col_idx]
                                ws.write(row_idx,col_idx,value,format_cache[col_idx])
                        elif rowtype == "wf_ponderation":
                            ws.write(row_idx,0,row["week"],ponderation_format)
                        elif rowtype == "wf_description":
                            ws.set_row(row_idx,60)
                            ws.write(row_idx,0,row["week"],ponderation_format)
                            ws.merge_range(row_idx,1,row_idx,len(df.columns)-3,row["outcomes"],description_text_format)
                        elif rowtype == "a_title" or rowtype == "b_description":
                            colour = row["colour"]
                            if pd.isna(colour) or colour == "":
                                continue
                            key = str(colour)+str(rowtype)

                            if key not in format_cache:
                                if rowtype == "a_title":
                                    format_cache[key] = workbook.add_format({
                                        "bg_color": colour,
                                        "border": 1,
                                        "font_color": readable_text_color(colour),
                                        "text_wrap": True,
                                        "valign": "top",
                                    })
                                else:
                                    format_cache[key] = workbook.add_format({
                                        "border": 1,
                                        "font_color": "#000000",
                                        "text_wrap": True,
                                        "valign": "top",
                                    })

                            cell_format = format_cache[key]
                            for col_idx in data_cols.values():
                                value = row.iloc[col_idx]
                                if pd.notna(value) and value != "":
                                    ws.write(row_idx, col_idx, value, cell_format)
        elif export_format == "csv":
            df = pd.DataFrame(
                {},
                columns=["week", "outcomes"],
            )
            for i, workflow in enumerate(workflows):
                df = pd.concat(
                    [df, get_workflow_full_table(workflow, allowed_sets)[0]]
                )
            df.to_csv(path_or_buf=b, sep=",", index=False)
        return b.getvalue()
